#include "petscksp.h"
#include "block.h"
#include "grid_jobs.h"
#include "invert.h"
#include "general.h"
#include "integration.h"
#include <iostream>
#include <complex>
#include <fstream>
#include <ctime>
using namespace std;

int main(int argc,char **args)
{
	//PETSc Declarations and Initializations
	PetscInitialize(&argc, &args);
	PetscInt size, rank;
	const PetscInt rows = 3600, columns = rows;
	PetscErrorCode ierr;
	Mat hamiltonian, inverseHamiltonian, identityMat;
	PC pc;
	KSP ksp;
	
	//Variable Declarations and Initializations
	block blockHamiltonian[300][3];
	time_t loopStart;
	double energyMin = -0.2, energyMax = -0.19985, energyStep = 0.00005, energy = energyMin,
	a = 2, atomCoordinates[rows][3], corrsd;
	const int leadSize = 4;
	int	isl[leadSize] = {1, 2, 3589, 3590},
		isr[leadSize] = {1789, 1790, 1801, 1802},
		isl180[leadSize] = {1, 2, 3589, 3590},
		isr180[leadSize] = {1801, 1802, 1789, 1790},
		vsdSteps = 100,
		IVcount = (energyMax >= 0 ? (energyMax - energyMin) / energyStep + .5
				: (energyMax - energyMin) / energyStep + 1.5);
	double IVvalues[IVcount][2], vsd, vsdMin = 0, vsdMax = .1, BField = 0.0; //(8./32)*(0.0255664/2)
	complex<double> gl[4][4], gr[4][4], dos2;
	
	//Output file
	ofstream outFile("PETScCntor3600.txt");
	if(!outFile.is_open())
	{
		ierr = PetscPrintf(PETSC_COMM_WORLD, "Error: Output file not opened!\n", size);CHKERRQ(ierr);
		return 0;
	} //endif
	
	//IVoutput file
	ofstream IVoutput("IVoutput.txt");
	if(!IVoutput.is_open())
	{
		ierr = PetscPrintf(PETSC_COMM_WORLD, "Error: IVoutput file not opened!\n", size);CHKERRQ(ierr);
		return 0;
	} //endif
	
//******************************************************************************************************
	//Checking number of processors
	ierr = MPI_Comm_size(PETSC_COMM_WORLD, &size);CHKERRQ(ierr);
	ierr = MPI_Comm_rank(PETSC_COMM_WORLD, &rank);CHKERRQ(ierr);
	ierr = PetscPrintf(PETSC_COMM_WORLD, "\nProcessors in use: %d\n", size);CHKERRQ(ierr);
	
	//Setup hamiltonian
	ierr = MatCreate(PETSC_COMM_WORLD, &hamiltonian);CHKERRQ(ierr);
	ierr = MatSetType(hamiltonian, MATAIJ);CHKERRQ(ierr);
	ierr = MatSetSizes(hamiltonian, PETSC_DECIDE, PETSC_DECIDE, rows, columns);CHKERRQ(ierr);
	
	//Setup PETSc Identity
	ierr = MatCreate(PETSC_COMM_WORLD, &identityMat);CHKERRQ(ierr);
	ierr = MatSetSizes(identityMat, PETSC_DECIDE, PETSC_DECIDE, rows, columns);CHKERRQ(ierr);
	ierr = MatSetType(identityMat, MATDENSE);CHKERRQ(ierr);
	for(int i=0; i<rows; i++)
		ierr = MatSetValue(identityMat, i, i, 1.0, INSERT_VALUES);CHKERRQ(ierr);
	ierr = MatAssemblyBegin(identityMat, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	ierr = MatAssemblyEnd(identityMat, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	
	//Setup PETSc inverseMat
	ierr = MatDuplicate(identityMat, MAT_DO_NOT_COPY_VALUES, &inverseHamiltonian);CHKERRQ(ierr);
	ierr = MatAssemblyBegin(inverseHamiltonian, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	ierr = MatAssemblyEnd(inverseHamiltonian, MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	
	//Setup PETSc Solver
	ierr = KSPCreate(PETSC_COMM_WORLD, &ksp);CHKERRQ(ierr);
	ierr = KSPSetOperators(ksp, hamiltonian, hamiltonian, SAME_NONZERO_PATTERN);CHKERRQ(ierr);
	ierr = KSPGetPC(ksp, &pc);CHKERRQ(ierr);
	ierr = PCSetType(pc, PCLU);CHKERRQ(ierr);
	ierr = KSPSetFromOptions(ksp);CHKERRQ(ierr);
	ierr = PCFactorSetZeroPivot(pc, 1e-18);CHKERRQ(ierr);
//******************************************************************************************************

	//Read in and setup system
	ierr = PetscPrintf(PETSC_COMM_WORLD, "Energy range: %.5f eV to %.5f eV\n", energyMin, energyMax);CHKERRQ(ierr);
	ierr = PetscPrintf(PETSC_COMM_WORLD, "BField = %f\n", BField);CHKERRQ(ierr);
	ierr = PetscPrintf(PETSC_COMM_WORLD, "\nEnergy\t\tDensityofStates\t   ScaledTransmissionFunction\tRe(fermif)\tTime\n\n");CHKERRQ(ierr);	
 	ReadInCoordinates(atomCoordinates);
	InitGridOffDiag(blockHamiltonian, energy, BField, atomCoordinates);
	
	//Loop over energy values
	for(energy=energyMin; energy<=energyMax; energy+=energyStep)
	{
		if(energy != 0)
		{
			time(&loopStart);
			InitGridDiag(blockHamiltonian, energy, BField, atomCoordinates);	
			wiregf(a, energy, 100, 100, isl180, isr180, isl, isr, gl, gr, atomCoordinates);
			AttachLeads(blockHamiltonian, isl180, isr180, isl, isr, gl, gr);
			ierr = InvertHamiltonian(hamiltonian, inverseHamiltonian, identityMat, pc, ksp, blockHamiltonian);CHKERRQ(ierr);
			ierr = RetrieveInformation(inverseHamiltonian, energy, energyMin, energyMax, energyStep, leadSize, IVvalues, outFile, isl, isr, gl, gr);CHKERRQ(ierr);
			DetachLeads(blockHamiltonian, isl180, isr180, isl, isr, gl, gr);
			ierr = PetscPrintf(PETSC_COMM_WORLD, "\t%.0f\n", difftime(time(NULL), loopStart));CHKERRQ(ierr);
		}
	} //endfor
	outFile.close();
	
	//Integrate IV curve
	//STILL UNCONFIRMED
	ierr = PetscPrintf(PETSC_COMM_WORLD, "\n");CHKERRQ(ierr);
	if(IVcount >= 6)
	{
		for(int i=0; i<=vsdSteps; i++)
		{
			vsd = vsdMin + i*(vsdMax-vsdMin)/vsdSteps;
			corrsd = integrate(IVvalues, IVcount, energyStep, vsd);
			ierr = PetscPrintf(PETSC_COMM_WORLD, "%.3f %.5e\n", vsd, corrsd);CHKERRQ(ierr);
			IVoutput << vsd << " " << corrsd << endl;
		} //endfor
	}
	else
	{
		char message[] = "IV Integration not done. Must complete at least 6 energy steps for integration method to work.\n";
		ierr = PetscPrintf(PETSC_COMM_WORLD, message);CHKERRQ(ierr);
		IVoutput << message;
	}
	IVoutput.close();
	
	//Destroy PETSc Objects
    ierr = MatDestroy(identityMat);CHKERRQ(ierr);
    ierr = MatDestroy(inverseHamiltonian);CHKERRQ(ierr);
    ierr = MatDestroy(hamiltonian);CHKERRQ(ierr);
	ierr = KSPDestroy(ksp);CHKERRQ(ierr);
	ierr = PetscFinalize();CHKERRQ(ierr);	
	return 0;
} //end main
